#include <stdio.h>
#include <suntool/sunview.h>
#include <suntool/canvas.h>
#include <suntool/panel.h>
#include "disp.h"

u_char		red[CMAPSIZE], green[CMAPSIZE], blue[CMAPSIZE];
Pixwin		*pw;

extern		rotate(), zoom();
extern float	transmat[4][4];

polygon	*polylist;

/*
 * simple program to display a polygon file
 */
main(ac, av)
	int	ac;
	char	**av;
{
	Frame		frame;
	Canvas		canvas;
	float		fmax, fmin;
	int		i, j, verts, connections, edges, npoints, pointno;
	Panel		panel;
	rotdata		xdata, ydata, zdata;
	vertex		*vlist;
	facet		*face;
	polygon		*poly;
	FILE		*mfile, *infile;
	char		name[80];
	float		xoff, yoff, zoff;

	if (ac != 2) {
		fprintf("usage: disp off_file_name\n");
		exit(1);
	}

	polylist = (polygon *)NULL;

	fmax = 1.0;
	fmin = -1.0;

	infile = fopen(av[1], "r");

	fscanf(infile, "%d %d %d\n", &verts, &connections, &edges);

	poly = (polygon *)malloc(sizeof(polygon));

	poly->nxt = polylist;
	polylist = poly;

	poly->pnts = verts;
	poly->vl = vlist = (vertex *)malloc(sizeof(vertex) * verts);
	poly->fl = (facet *)NULL;

	for (i = 0; i != verts; i++) {
		fscanf(infile, "%f %f %f\n", &vlist[i].x, &vlist[i].y, &vlist[i].z);
		vlist[i].x += xoff;
		vlist[i].y += yoff;
		vlist[i].z += zoff;

		if (vlist[i].x < fmin)
			fmin = vlist[i].x;
		else if (vlist[i].x > fmax)
			fmax = vlist[i].x;
		if (vlist[i].y < fmin)
			fmin = vlist[i].y;
		else if (vlist[i].y > fmax)
			fmax = vlist[i].y;
		if (vlist[i].z < fmin)
			fmin = vlist[i].z;
		else if (vlist[i].z > fmax)
			fmax = vlist[i].z;
	}

	for (i = 0; i != connections; i++) {
		fscanf(infile, "%d", &npoints);
		face = (facet *)malloc(sizeof(facet));
		face->verts = (vertex **)malloc((npoints + 1) * sizeof(vertex **));
		for (j = 0; j != npoints; j++) {
			fscanf(infile, "%d", &pointno);
			face->verts[j] = &vlist[pointno - 1];
		}
		face->verts[npoints] = face->verts[0];
		face->next = poly->fl;
		poly->fl = face;
	}

	fclose(infile);

	if (fmax < -fmin)
		transmat[3][3] /= -fmin;
	else
		transmat[3][3] /= fmax;

	frame = window_create(
			0, FRAME,
			WIN_X, 0,
			WIN_Y, 0,
			FRAME_LABEL, av[1],
		0);

	canvas = window_create(
			frame, CANVAS,
			WIN_HEIGHT, SIZE,
			WIN_WIDTH, SIZE,
			CANVAS_RETAINED, FALSE,
		0);

	panel = window_create(
			frame, PANEL, 
			WIN_BELOW, canvas,
			WIN_X, 0,
		0);

	red[0] = green[0] = blue[0] = 0;
	red[1] = 255;
	green[2] = 255;
	red[3] = green[3] = 255;
	blue[4] = 255;
	red[5] = blue[5] = 255;
	green[6] = blue[6] = 255;
	red[7] = green[7] = blue[7] = 255;

	for (i = 8; i < CMAPSIZE; i++)
		red[i] = green[i] = blue[i] = 255;

	pw = (Pixwin *)window_get(frame, WIN_PIXWIN);
	pw_setcmsname(pw, "disp");
	pw_putcolormap(pw, 0, CMAPSIZE, red, green, blue);

	pw = (Pixwin *)window_get(panel, WIN_PIXWIN);
	pw_setcmsname(pw, "disp");
	pw_putcolormap(pw, 0, CMAPSIZE, red, green, blue);

	pw = canvas_pixwin(canvas);
	pw_setcmsname(pw, "disp");
	pw_putcolormap(pw, 0, CMAPSIZE, red, green, blue);

	mident4(xdata.m);
	xdata.type = X;
	xdata.oldv = 0.0;
	mident4(ydata.m);
	ydata.type = Y;
	ydata.oldv = 0.0;
	mident4(zdata.m);
	zdata.type = Z;
	zdata.oldv = 0.0;
	
	panel_create_item(
		panel, PANEL_SLIDER,
		PANEL_LABEL_STRING, "x rot: ",
		PANEL_VALUE, 0,
		PANEL_MIN_VALUE, -180,
		PANEL_MAX_VALUE, 180,
		PANEL_SLIDER_WIDTH, 360,
		PANEL_CLIENT_DATA, &xdata,
		PANEL_NOTIFY_LEVEL, PANEL_ALL,
		PANEL_NOTIFY_PROC, rotate,
	0);

	panel_create_item(
		panel, PANEL_SLIDER,
		PANEL_LABEL_STRING, "y rot: ",
		PANEL_VALUE, 0,
		PANEL_MIN_VALUE, -180,
		PANEL_MAX_VALUE, 180,
		PANEL_SLIDER_WIDTH, 360,
		PANEL_CLIENT_DATA, &ydata,
		PANEL_NOTIFY_LEVEL, PANEL_ALL,
		PANEL_NOTIFY_PROC, rotate,
	0);

	panel_create_item(
		panel, PANEL_SLIDER,
		PANEL_LABEL_STRING, "z rot: ",
		PANEL_VALUE, 0,
		PANEL_MIN_VALUE, -180,
		PANEL_MAX_VALUE, 180,
		PANEL_SLIDER_WIDTH, 360,
		PANEL_CLIENT_DATA, &zdata,
		PANEL_NOTIFY_LEVEL, PANEL_ALL,
		PANEL_NOTIFY_PROC, rotate,
	0);

	window_set(canvas,
		CANVAS_RETAINED, TRUE,
		0);

	window_fit(canvas);
	window_fit(panel);
	window_fit(frame);

	pw = canvas_pixwin(canvas);

	pw_writebackground(pw, 0, 0, SIZE, SIZE, PIX_SRC);

	for (poly = polylist; poly != (polygon *)NULL; poly = poly->nxt)
		draw(poly);

	window_main_loop(frame);
	exit(0);
}
